const fs = require('fs');
const path = require('path');
const camelCase = require('lodash/camelCase');

const filenameToInterfaceName = (filename) => {
    return path.basename(filename)
        .replace(/^(\w)/, (_, c) => 'I' + c.toUpperCase())
        .replace(/\W+(\w)/g, (_, c) => c.toUpperCase());
};

const cssModuleToTypescriptInterfaceProperties = (cssModuleKeys, indent = '  ') => {
    return cssModuleKeys
        .map((key) => `${indent}'${key}': string;`)
        .join('\n');
};

const cssModuleToNamedExports = (cssModuleKeys) => {
    return cssModuleKeys
        .map((key) => `export const ${key}: string;`)
        .join('\n');
};

const allWordsRegexp = /^\w+$/i;

const filterNonWordClasses = (cssModuleKeys) => {
    const filteredClassNames = cssModuleKeys.filter(classname => allWordsRegexp.test(classname));
    if (filteredClassNames.length === cssModuleKeys.length) {
        return [filteredClassNames, [],];
    }
    const nonWordClassNames = cssModuleKeys.filter(classname => !allWordsRegexp.test(classname));
    return [filteredClassNames, nonWordClassNames,];
};

// Documented here: https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Lexical_grammar#Reserved_keywords_as_of_ECMAScript_2015
const reservedWords = [
    'break',
    'case',
    'catch',
    'class',
    'const',
    'continue',
    'debugger',
    'default',
    'delete',
    'do',
    'else',
    'export',
    'extends',
    'finally',
    'for',
    'function',
    'if',
    'import',
    'in',
    'instanceof',
    'new',
    'return',
    'super',
    'switch',
    'this',
    'throw',
    'try',
    'typeof',
    'var',
    'void',
    'while',
    'with',
    'yield'
];

const filterReservedWordClasses = (cssModuleKeys) => {
    const filteredClassNames = cssModuleKeys.filter(classname => reservedWords.indexOf(classname) === -1);
    if (filteredClassNames.length === cssModuleKeys.length) {
        return [filteredClassNames, [],];
    }
    const reservedWordClassNames = cssModuleKeys.filter(classname => reservedWords.indexOf(classname) !== -1);
    return [filteredClassNames, reservedWordClassNames,];
};


const filenameToTypingsFilename = (filename) => {
    const dirName = path.dirname(filename);
    const baseName = path.basename(filename);
    return path.join(dirName, `${baseName}.d.ts`);
};

const generateNamedExports = (cssModuleKeys) => {
    const namedExports = cssModuleToNamedExports(cssModuleKeys);
    return (`${namedExports}
`);
};

const generateGenericExportInterface = (cssModuleKeys, filename, indent) => {
    const interfaceName = filenameToInterfaceName(filename);
    const interfaceProperties = cssModuleToTypescriptInterfaceProperties(cssModuleKeys, indent);
    return (
        `export interface ${interfaceName} {
${interfaceProperties}
}

export const locals: ${interfaceName};
`);
};

const generateCssTypings = (filePath, options = {}) => {
    const filename = filePath;
    const cssModuleInterfaceFilename = filenameToTypingsFilename(filename);

    const keyRegex = /([_a-zA-Z]+[_a-zA-Z0-9-]*)\s*{/g;
    let match;
    let cssModuleKeys = [];
    const content = fs.readFileSync(filePath);

    while (match = keyRegex.exec(content.toString())) {
        if (cssModuleKeys.indexOf(match[1]) < 0) {
            cssModuleKeys.push(match[1]);
        }
    }

    if (options.camelCase) {
        cssModuleKeys = cssModuleKeys.map(key => camelCase(key));
    }

    let cssModuleDefinition;

    if (!options.namedExport) {
        cssModuleDefinition = generateGenericExportInterface(cssModuleKeys, filename);
    } else {
        const [cleanedDefinitions, skippedDefinitions,] = filterNonWordClasses(cssModuleKeys);
        if (skippedDefinitions.length > 0 && !options.camelCase) {
            console.warn('warn', `Typings for CSS-Modules: option 'namedExport' was set but 'camelCase' for the css-loader not.
The following classes will not be available as named exports:
${skippedDefinitions.map(sd => ` - "${sd}"`).join('\n').red}
`.yellow);
        }

        const [nonReservedWordDefinitions, reservedWordDefinitions,] = filterReservedWordClasses(cleanedDefinitions);
        if (reservedWordDefinitions.length > 0) {
            console.warn('warn', `Your css contains classes which are reserved words in JavaScript.
Consequently the following classes will not be available as named exports:
${reservedWordDefinitions.map(rwd => ` - "${rwd}"`).join('\n').red}
These can be accessed using the object literal syntax; eg styles['delete'] instead of styles.delete.
`.yellow);
        }

        cssModuleDefinition = generateNamedExports(nonReservedWordDefinitions);
    }

    if (cssModuleDefinition.trim() === '') {
        // Ensure empty CSS modules export something
        cssModuleDefinition = 'export {};\n';
    }

    if (options.banner) {
        // Prefix banner to CSS module
        cssModuleDefinition = options.banner + '\n' + cssModuleDefinition;
    }

    fs.writeFileSync(cssModuleInterfaceFilename, cssModuleDefinition);
};

module.exports = {
    generateCssTypings,
};